package activitypub

import (
	"encoding/json"
	"fmt"
	"net/http"
	"strconv"
	"time"

	webutils "git.mstar.dev/mstar/goutils/http"
	"git.mstar.dev/mstar/goutils/other"
	"git.mstar.dev/mstar/goutils/sliceutils"
	"github.com/rs/zerolog/hlog"
	"gorm.io/gorm"

	"git.mstar.dev/mstar/linstrom/activitypub"
	"git.mstar.dev/mstar/linstrom/config"
	"git.mstar.dev/mstar/linstrom/shared"
	"git.mstar.dev/mstar/linstrom/storage-new"
	"git.mstar.dev/mstar/linstrom/storage-new/dbgen"
	"git.mstar.dev/mstar/linstrom/storage-new/models"
	webshared "git.mstar.dev/mstar/linstrom/web/shared"
)

func users(w http.ResponseWriter, r *http.Request) {
	type OutboundKey struct {
		Id    string `json:"id"`
		Owner string `json:"owner"`
		Pem   string `json:"publicKeyPem"`
	}
	type OutboundMedia struct {
		Type      string `json:"type"`
		Url       string `json:"url"`
		MediaType string `json:"mediaType"`
	}
	type Outbound struct {
		Context           []any          `json:"@context"`
		Id                string         `json:"id"`
		Type              string         `json:"type"`
		PreferredUsername string         `json:"preferredUsername"`
		Inbox             string         `json:"inbox"`
		Outboux           string         `json:"outbox"`
		PublicKey         OutboundKey    `json:"publicKey"`
		Published         time.Time      `json:"published"`
		DisplayName       string         `json:"name"`
		Description       *string        `json:"summary,omitempty"`
		PublicUrl         string         `json:"url"`
		Icon              *OutboundMedia `json:"icon,omitempty"`
		Banner            *OutboundMedia `json:"image,omitempty"`
		Discoverable      bool           `json:"discoverable"`
		Location          *string        `json:"vcard:Address,omitempty"`
		Birthday          *string        `json:"vcard:bday,omitempty"`
		SpeakAsCat        bool           `json:"speakAsCat"`
		IsCat             bool           `json:"isCat"`
		RestrictedFollow  bool           `json:"manuallyApprovesFollowers"`
		Following         string         `json:"following"`
		Followers         string         `json:"followers"`
	}
	log := hlog.FromRequest(r)
	userId := r.PathValue("id")
	user, err := dbgen.User.Where(dbgen.User.ID.Eq(userId)).
		Preload(dbgen.User.Icon).Preload(dbgen.User.Banner).
		Preload(dbgen.User.BeingTypes).
		First()
	if err != nil {
		_ = webutils.ProblemDetails(
			w,
			500,
			"/errors/db-failure",
			"internal database failure",
			nil,
			nil,
		)
		if storage.HandleReconnectError(err) {
			log.Warn().Msg("Connection to db lost. Reconnect attempt started")
		} else {
			log.Error().Err(err).Msg("Failed to get total user count from db")
		}
		return
	}
	// FIXME: Remove this later
	// (or rather move to dedicated module in storage for old migration stuff),
	// temporary fix for old data. User creation locations are fixed already
	err = storage.EnsureLocalUserIdHasLinks(userId)
	if err != nil {
		log.Warn().Err(err).Msg("Failed to create links for local user")
	}

	apUrl := activitypub.UserIdToApUrl(user.ID)
	var keyBytes string
	if config.GlobalConfig.Experimental.UseEd25519Keys {
		keyBytes = shared.KeyBytesToPem(user.PublicKeyEd, true)
	} else {
		keyBytes = shared.KeyBytesToPem(user.PublicKeyRsa, false)
	}
	data := Outbound{
		Context:           activitypub.BaseLdContext,
		Id:                apUrl,
		Type:              "Person",
		PreferredUsername: user.Username,
		Inbox:             apUrl + "/inbox",
		Outboux:           apUrl + "/outbox",
		PublicKey: OutboundKey{
			Id:    apUrl + "#main-key",
			Owner: apUrl,
			Pem:   keyBytes,
		},
		Published:        user.CreatedAt,
		DisplayName:      user.DisplayName,
		PublicUrl:        config.GlobalConfig.General.GetFullPublicUrl() + "/user/" + user.Username,
		Discoverable:     user.Indexable,
		RestrictedFollow: user.RestrictedFollow,
		Following:        apUrl + "/following",
		Followers:        apUrl + "/followers",
	}
	if user.Description != "" {
		data.Description = &user.Description
	}
	if user.Icon != nil {
		log.Debug().Msg("icon found")
		data.Icon = &OutboundMedia{
			Type: "Image",
			Url: config.GlobalConfig.General.GetFullPublicUrl() + webshared.EnsurePublicUrl(
				user.Icon.Location,
			),
			MediaType: user.Icon.Type,
		}
	}
	if user.Banner != nil {
		log.Debug().Msg("icon banner")
		data.Banner = &OutboundMedia{
			Type: "Image",
			Url: config.GlobalConfig.General.GetFullPublicUrl() + webshared.EnsurePublicUrl(
				user.Banner.Location,
			),
			MediaType: user.Banner.Type,
		}
	}
	if sliceutils.ContainsFunc(user.BeingTypes, func(t models.UserToBeing) bool {
		return t.Being == string(models.BEING_CAT)
	}) {
		data.IsCat = true
		// data.SpeakAsCat = true // TODO: Move to check of separate field in db model
	}
	if user.Location.Valid {
		data.Location = &user.Location.String
	}
	if user.Birthday.Valid {
		data.Birthday = &user.Birthday.String
		// data.Birthday = other.IntoPointer(user.Birthday.Time.Format("2006-Jan-02")) //YYYY-Month-DD
	}

	encoded, err := json.Marshal(data)
	if err != nil {
		log.Error().Err(err).Msg("Failed to marshal response")
		_ = webutils.ProblemDetailsStatusOnly(w, http.StatusInternalServerError)
		return
	}
	w.Header().Add("Content-Type", "application/activity+json")
	_, _ = fmt.Fprint(w, string(encoded))
}

func userFollowing(w http.ResponseWriter, r *http.Request) {
	log := hlog.FromRequest(r)
	userId := r.PathValue("id")
	pageNrStr := r.FormValue("page")
	exists, err := dbgen.User.DoesUserWithIdExist(userId)
	if err != nil {
		log.Error().Err(err).Str("id", userId).Msg("Failed to check if user exists")
		_ = webutils.ProblemDetailsStatusOnly(w, http.StatusInternalServerError)
		return
	}
	if !exists {
		_ = webutils.ProblemDetailsStatusOnly(w, http.StatusNotFound)
		return
	}
	apUrl := activitypub.UserIdToApUrl(userId)
	var data []byte
	followingCount, err := dbgen.UserToUserRelation.CountFollowingForId(userId)
	if err != nil {
		log.Error().Err(err).Str("id", userId).Msg("Failed to get following count")
		_ = webutils.ProblemDetailsStatusOnly(w, http.StatusInternalServerError)
		return
	}
	if pageNrStr == "" {
		col := collectionOut{
			Context:    "https://www.w3.org/ns/activitystreams",
			Type:       "OrderedCollection",
			Id:         apUrl + "/following",
			TotalItems: followingCount,
			First:      apUrl + "/following?page=0",
		}
		data, err = json.Marshal(col)
		if err != nil {
			log.Error().Err(err).Any("raw", data).Msg("Failed to marshal following collection page")
			_ = webutils.ProblemDetailsStatusOnly(w, http.StatusInternalServerError)
			return
		}
	} else {
		pageNr, err := strconv.Atoi(pageNrStr)
		if err != nil {
			_ = webutils.ProblemDetails(
				w,
				http.StatusBadRequest,
				"/errors/bad-page",
				"bad page number",
				other.IntoPointer("page number must be an uint"),
				nil,
			)
			return
		}
		hasNextPage := followingCount-(pageNr+1)*50 > 0
		hasPreviousPage := pageNr > 0
		links, err := dbgen.UserToUserRelation.GetFollowingApLinksPagedForId(userId, pageNr)
		switch err {
		case gorm.ErrRecordNotFound:
			_ = webutils.ProblemDetailsStatusOnly(w, http.StatusNotFound)
			return
		case nil:
		default:
			log.Error().Err(err).Str("id", userId).Msg("Failed to get account via id")
			_ = webutils.ProblemDetailsStatusOnly(w, http.StatusInternalServerError)
			return
		}
		page := collectionPageOut{
			Context: "https://www.w3.org/ns/activitystreams",
			Type:    "OrderedCollectionPage",
			Id:      fmt.Sprintf("%s/following?page=%d", apUrl, pageNr),
			PartOf:  activitypub.UserIdToApUrl(userId) + "/following",
			Items:   sliceutils.Map(links, func(t string) any { return t }),
		}
		if hasNextPage {
			page.Next = fmt.Sprintf("%s/following?page=%d", apUrl, pageNr+1)
		}
		if hasPreviousPage {
			page.Next = fmt.Sprintf("%s/following?page=%d", apUrl, pageNr-1)
		}
		data, err = json.Marshal(page)
		if err != nil {
			log.Error().Err(err).Any("raw", page).Msg("Failed to marshal following collection page")
			_ = webutils.ProblemDetailsStatusOnly(w, http.StatusInternalServerError)
			return
		}
	}
	w.Header().Add("Content-Type", "application/activity+json")
	_, _ = fmt.Fprint(w, string(data))
}

func userFollowers(w http.ResponseWriter, r *http.Request) {
	log := hlog.FromRequest(r)
	userId := r.PathValue("id")
	pageNrStr := r.FormValue("page")
	exists, err := dbgen.User.DoesUserWithIdExist(userId)
	if err != nil {
		log.Error().Err(err).Str("id", userId).Msg("Failed to check if user exists")
		_ = webutils.ProblemDetailsStatusOnly(w, http.StatusInternalServerError)
		return
	}
	if !exists {
		log.Debug().Str("id", userId).Msg("user not found")
		_ = webutils.ProblemDetailsStatusOnly(w, http.StatusNotFound)
		return
	}
	apUrl := activitypub.UserIdToApUrl(userId)
	var data []byte
	followersCount, err := dbgen.UserToUserRelation.CountFollowersForId(userId)
	if err != nil {
		log.Error().Err(err).Str("id", userId).Msg("Failed to get followers count")
		_ = webutils.ProblemDetailsStatusOnly(w, http.StatusInternalServerError)
		return
	}
	if pageNrStr == "" {
		col := collectionOut{
			Context:    activitypub.BaseLdContext,
			Type:       "OrderedCollection",
			Id:         apUrl + "/followers",
			TotalItems: followersCount,
			First:      apUrl + "/followers?page=0",
		}
		data, err = json.Marshal(col)
		if err != nil {
			log.Error().Err(err).Any("raw", data).Msg("Failed to marshal followers collection page")
			_ = webutils.ProblemDetailsStatusOnly(w, http.StatusInternalServerError)
			return
		}
	} else {
		pageNr, err := strconv.Atoi(pageNrStr)
		if err != nil {
			_ = webutils.ProblemDetails(
				w,
				http.StatusBadRequest,
				"/errors/bad-page",
				"bad page number",
				other.IntoPointer("page number must be an uint"),
				nil,
			)
			return
		}
		hasNextPage := followersCount-(pageNr+1)*50 > 0
		hasPreviousPage := pageNr > 0
		links, err := dbgen.UserToUserRelation.GetFollowerApLinksPagedForId(userId, pageNr)
		switch err {
		case gorm.ErrRecordNotFound:
			log.Debug().Msg("No followers found")
			_ = webutils.ProblemDetailsStatusOnly(w, http.StatusNotFound)
			return
		case nil:
		default:
			log.Error().Err(err).Str("id", userId).Msg("Failed to get account via id")
			_ = webutils.ProblemDetailsStatusOnly(w, http.StatusInternalServerError)
			return
		}
		page := collectionPageOut{
			Context: activitypub.BaseLdContext,
			Type:    "OrderedCollectionPage",
			Id:      fmt.Sprintf("%s/followers?page=%d", apUrl, pageNr),
			PartOf:  activitypub.UserIdToApUrl(userId) + "/followers",
			Items:   sliceutils.Map(links, func(t string) any { return t }),
		}
		if hasNextPage {
			page.Next = fmt.Sprintf("%s/followers?page=%d", apUrl, pageNr+1)
		}
		if hasPreviousPage {
			page.Next = fmt.Sprintf("%s/followers?page=%d", apUrl, pageNr-1)
		}
		data, err = json.Marshal(page)
		if err != nil {
			log.Error().Err(err).Any("raw", page).Msg("Failed to marshal followers collection page")
			_ = webutils.ProblemDetailsStatusOnly(w, http.StatusInternalServerError)
			return
		}
	}
	log.Debug().Bytes("body", data).Msg("Sending collection(page) out")
	w.Header().Add("Content-Type", "application/activity+json")
	_, _ = fmt.Fprint(w, string(data))
}

/*
Fine. You win JsonLD. I can't get you to work properly. I'll just treat you like normal json then
Fuck you.

If anyone wants to get this shit working *the propper way* with JsonLD, here's the
original code
var chain goap.BaseApChain = &goap.EmptyBaseObject{}
	chain = goap.AppendUDIdData(chain, apUrl)
	chain = goap.AppendUDTypeData(chain, "Person")
	chain = goap.AppendASPreferredNameData(chain, goap.ValueValue[string]{Value: user.DisplayName})
	chain = goap.AppendW3SecurityPublicKeyData(
		chain,
		apUrl+"#main-key",
		apUrl,
		keyBytesToPem(user.PublicKey),
	)

	chainMap := chain.MarshalToMap()
	proc := ld.NewJsonLdProcessor()
	options := ld.NewJsonLdOptions("")
	tmp, tmperr := json.Marshal(chainMap)
	fmt.Println(string(tmp), tmperr)
	data, err := proc.Compact(chainMap, baseLdContext, options)
	// data, err := goap.Compact(chain, baseLdContext)
	if err != nil {
		log.Error().Err(err).Msg("Failed to marshal ap chain")
		_ = webutils.ProblemDetailsStatusOnly(w, 500)
		return
	}
*/
